@@ -18,23 +18,241 @@
 #include <thread>
 #include <chrono>
 
+#include "rcutils/error_handling.h"
+#include "rcutils/types.h"
+
+#include "rcpputils/scope_exit.hpp"
+
 #include "rmw/allocators.h"
 #include "rmw/error_handling.h"
-#include "rmw/rmw.h"
 #include "rmw/impl/cpp/macros.hpp"
 #include "rmw/types.h"
+#include "rmw/validate_full_topic_name.h"
 
-#include "rmw_gurumdds_shared_cpp/rmw_common.hpp"
-#include "rmw_gurumdds_shared_cpp/types.hpp"
-#include "rmw_gurumdds_shared_cpp/qos.hpp"
-#include "rmw_gurumdds_shared_cpp/namespace_prefix.hpp"
+#include "rmw_gurumdds_cpp/gid.hpp"
+#include "rmw_gurumdds_cpp/graph_cache.hpp"
 #include "rmw_gurumdds_cpp/identifier.hpp"
+#include "rmw_gurumdds_cpp/namespace_prefix.hpp"
+#include "rmw_gurumdds_cpp/names_and_types_helpers.hpp"
+#include "rmw_gurumdds_cpp/qos.hpp"
+#include "rmw_gurumdds_cpp/rmw_context_impl.hpp"
+#include "rmw_gurumdds_cpp/rmw_publisher.hpp"
 #include "rmw_gurumdds_cpp/types.hpp"
 
-#include "rcutils/types.h"
-#include "rcutils/error_handling.h"
+rmw_publisher_t *
+__rmw_create_publisher(
+  rmw_context_impl_t * const ctx,
+  const rmw_node_t * node,
+  dds_DomainParticipant * const participant,
+  dds_Publisher * const pub,
+  const rosidl_message_type_support_t * type_supports,
+  const char * topic_name,
+  const rmw_qos_profile_t * qos_policies,
+  const rmw_publisher_options_t * publisher_options,
+  const bool internal)
+{
+  std::lock_guard<std::mutex> guard(ctx->endpoint_mutex);
 
-#include "./type_support_common.hpp"
+  const rosidl_message_type_support_t * type_support =
+    get_message_typesupport_handle(type_supports, rosidl_typesupport_introspection_c__identifier);
+  if (type_support == nullptr) {
+    rcutils_reset_error();
+    type_support = get_message_typesupport_handle(
+      type_supports, rosidl_typesupport_introspection_cpp::typesupport_identifier);
+    if (type_support == nullptr) {
+      rcutils_reset_error();
+      RMW_SET_ERROR_MSG("type support not from this implementation");
+      return nullptr;
+    }
+  }
+
+  rmw_publisher_t * rmw_publisher = nullptr;
+  GurumddsPublisherInfo * publisher_info = nullptr;
+  dds_DataWriter * topic_writer = nullptr;
+  dds_DataWriterQos datawriter_qos;
+  dds_Topic * topic = nullptr;
+  dds_TopicDescription * topic_desc = nullptr;
+  dds_TypeSupport * dds_typesupport = nullptr;
+  dds_ReturnCode_t ret;
+
+  std::string type_name =
+    create_type_name(type_support->data, type_support->typesupport_identifier);
+  if (type_name.empty()) {
+    // Error message is already set
+    return nullptr;
+  }
+
+  std::string processed_topic_name = create_topic_name(
+    ros_topic_prefix, topic_name, "", qos_policies);
+
+  std::string metastring =
+    create_metastring(type_support->data, type_support->typesupport_identifier);
+  if (metastring.empty()) {
+    // Error message is already set
+    return nullptr;
+  }
+
+  dds_typesupport = dds_TypeSupport_create(metastring.c_str());
+  if (dds_typesupport == nullptr) {
+    RMW_SET_ERROR_MSG("failed to create typesupport");
+    return nullptr;
+  }
+
+  ret = dds_TypeSupport_register_type(dds_typesupport, participant, type_name.c_str());
+  if (ret != dds_RETCODE_OK) {
+    RMW_SET_ERROR_MSG("failed to register type to domain participant");
+    return nullptr;
+  }
+
+  topic_desc = dds_DomainParticipant_lookup_topicdescription(
+    participant, processed_topic_name.c_str());
+  if (topic_desc == nullptr) {
+    dds_TopicQos topic_qos;
+    ret = dds_DomainParticipant_get_default_topic_qos(participant, &topic_qos);
+    if (ret != dds_RETCODE_OK) {
+      RMW_SET_ERROR_MSG("failed to get default topic qos");
+      return nullptr;
+    }
+
+    topic = dds_DomainParticipant_create_topic(
+      participant, processed_topic_name.c_str(), type_name.c_str(), &topic_qos, nullptr, 0);
+    if (topic == nullptr) {
+      RMW_SET_ERROR_MSG("failed to create topic");
+      dds_TopicQos_finalize(&topic_qos);
+      return nullptr;
+    }
+
+    ret = dds_TopicQos_finalize(&topic_qos);
+    if (ret != dds_RETCODE_OK) {
+      RMW_SET_ERROR_MSG("failed to finalize topic qos");
+      return nullptr;
+    }
+  } else {
+    dds_Duration_t timeout;
+    timeout.sec = 0;
+    timeout.nanosec = 1;
+    topic = dds_DomainParticipant_find_topic(participant, processed_topic_name.c_str(), &timeout);
+    if (topic == nullptr) {
+      RMW_SET_ERROR_MSG("failed to find topic");
+      return nullptr;
+    }
+  }
+
+  if (!get_datawriter_qos(pub, qos_policies, &datawriter_qos)) {
+    // Error message already set
+    return nullptr;
+  }
+
+  topic_writer = dds_Publisher_create_datawriter(pub, topic, &datawriter_qos, nullptr, 0);
+  if (topic_writer == nullptr) {
+    RMW_SET_ERROR_MSG("failed to create datawriter");
+    dds_DataWriterQos_finalize(&datawriter_qos);
+    return nullptr;
+  }
+
+  ret = dds_DataWriterQos_finalize(&datawriter_qos);
+  if (ret != dds_RETCODE_OK) {
+    RMW_SET_ERROR_MSG("failed to finalize datawriter qos");
+    return nullptr;
+  }
+
+  publisher_info = new(std::nothrow) GurumddsPublisherInfo();
+  if (publisher_info == nullptr) {
+    RMW_SET_ERROR_MSG("failed to allocate GurumddsPublisherInfo");
+    return nullptr;
+  }
+
+  publisher_info->topic_writer = topic_writer;
+  publisher_info->rosidl_message_typesupport = type_support;
+  publisher_info->implementation_identifier = RMW_GURUMDDS_ID;
+  publisher_info->sequence_number = 0;
+  publisher_info->ctx = ctx;
+
+  entity_get_gid(
+    reinterpret_cast<dds_Entity *>(publisher_info->topic_writer),
+    publisher_info->publisher_gid);
+
+  rmw_publisher = rmw_publisher_allocate();
+  if (rmw_publisher == nullptr) {
+    RMW_SET_ERROR_MSG("failed to allocate publisher");
+    return nullptr;
+  }
+  rmw_publisher->topic_name = nullptr;
+
+  auto scope_exit_rmw_publisher_delete = rcpputils::make_scope_exit(
+    [rmw_publisher]() {
+      if (rmw_publisher->topic_name != nullptr) {
+        rmw_free(const_cast<char *>(rmw_publisher->topic_name));
+      }
+      rmw_publisher_free(rmw_publisher);
+    });
+
+  rmw_publisher->implementation_identifier = RMW_GURUMDDS_ID;
+  rmw_publisher->data = publisher_info;
+  rmw_publisher->topic_name =
+    reinterpret_cast<const char *>(rmw_allocate(strlen(topic_name) + 1));
+  if (rmw_publisher->topic_name == nullptr) {
+    RCUTILS_LOG_ERROR_NAMED(RMW_GURUMDDS_ID, "failed to allocate publisher's topic name");
+    return nullptr;
+  }
+  memcpy(
+    const_cast<char *>(rmw_publisher->topic_name),
+    topic_name,
+    strlen(topic_name) + 1);
+  rmw_publisher->options = *publisher_options;
+  rmw_publisher->can_loan_messages = false;
+
+  if (!internal) {
+    if (graph_on_publisher_created(ctx, node, publisher_info) != RMW_RET_OK) {
+      RCUTILS_LOG_ERROR_NAMED(RMW_GURUMDDS_ID, "failed to update graph for publisher");
+      return nullptr;
+    }
+  }
+
+  dds_TypeSupport_delete(dds_typesupport);
+  dds_typesupport = nullptr;
+
+  scope_exit_rmw_publisher_delete.cancel();
+  return rmw_publisher;
+}
+
+rmw_ret_t
+__rmw_destroy_publisher(
+  rmw_context_impl_t * const ctx,
+  rmw_publisher_t * const publisher)
+{
+  std::lock_guard<std::mutex> guard(ctx->endpoint_mutex);
+
+  auto publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
+  if (publisher_info == nullptr) {
+    RMW_SET_ERROR_MSG("invalid publisher data");
+    return RMW_RET_ERROR;
+  }
+
+  dds_ReturnCode_t ret;
+  if (publisher_info->topic_writer != nullptr) {
+    dds_Topic * topic = dds_DataWriter_get_topic(publisher_info->topic_writer);
+    ret = dds_Publisher_delete_datawriter(ctx->publisher, publisher_info->topic_writer);
+    if (ret != dds_RETCODE_OK) {
+      RMW_SET_ERROR_MSG("failed to delete datawriter");
+      return RMW_RET_ERROR;
+    }
+    publisher_info->topic_writer = nullptr;
+
+    ret = dds_DomainParticipant_delete_topic(ctx->participant, topic);
+    if (ret == dds_RETCODE_PRECONDITION_NOT_MET) {
+      RCUTILS_LOG_DEBUG_NAMED(RMW_GURUMDDS_ID, "The entity using the topic still exists.");
+    } else if (ret != dds_RETCODE_OK) {
+      RMW_SET_ERROR_MSG("failed to delete topic");
+      return RMW_RET_ERROR;
+    }
+  }
+
+  delete publisher_info;
+  publisher->data = nullptr;
+
+  return RMW_RET_OK;
+}
 
 extern "C"
 {
@@ -65,267 +283,72 @@ rmw_publisher_t *
 rmw_create_publisher(
   const rmw_node_t * node,
   const rosidl_message_type_support_t * type_supports,
-  const char * topic_name, const rmw_qos_profile_t * qos_policies,
+  const char * topic_name,
+  const rmw_qos_profile_t * qos_policies,
   const rmw_publisher_options_t * publisher_options)
 {
-  if (node == nullptr) {
-    RMW_SET_ERROR_MSG("node handle is null");
+  RMW_CHECK_ARGUMENT_FOR_NULL(node, nullptr);
+  RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
+    node,
+    node->implementation_identifier,
+    RMW_GURUMDDS_ID,
+    return nullptr);
+  RMW_CHECK_ARGUMENT_FOR_NULL(type_supports, nullptr);
+  RMW_CHECK_ARGUMENT_FOR_NULL(topic_name, nullptr);
+  if (strlen(topic_name) == 0) {
+    RMW_SET_ERROR_MSG("topic_name argument is empty");
     return nullptr;
   }
+  RMW_CHECK_ARGUMENT_FOR_NULL(qos_policies, nullptr);
+  RMW_CHECK_ARGUMENT_FOR_NULL(publisher_options, nullptr);
 
-  if (node->implementation_identifier != gurum_gurumdds_identifier) {
-    RMW_SET_ERROR_MSG("node handle not from this implementation");
-    return nullptr;
-  }
-
-  if (topic_name == nullptr || strlen(topic_name) == 0) {
-    RMW_SET_ERROR_MSG("publisher topic is null or empty string");
-    return nullptr;
-  }
-
-  if (qos_policies == nullptr) {
-    RMW_SET_ERROR_MSG("qos profile is null");
-    return nullptr;
-  }
-
-  if (publisher_options == nullptr) {
-    RMW_SET_ERROR_MSG("publisher_options is null");
-    return nullptr;
+  if (!qos_policies->avoid_ros_namespace_conventions) {
+    int validation_result = RMW_TOPIC_VALID;
+    rmw_ret_t ret = rmw_validate_full_topic_name(topic_name, &validation_result, nullptr);
+    if (ret != RMW_RET_OK) {
+      return nullptr;
+    }
+    if (validation_result != RMW_TOPIC_VALID) {
+      const char * reason = rmw_full_topic_name_validation_result_string(validation_result);
+      RMW_SET_ERROR_MSG_WITH_FORMAT_STRING(
+        "topic name is invalid: %s", reason);
+      return nullptr;
+    }
   }
 
   if (publisher_options->require_unique_network_flow_endpoints ==
     RMW_UNIQUE_NETWORK_FLOW_ENDPOINTS_STRICTLY_REQUIRED)
   {
-    RMW_SET_ERROR_MSG(
-      "Strict requirement on unique network flow endpoints for publishers not supported");
+    RMW_SET_ERROR_MSG("Unique network flow endpoints not supported on publishers");
     return nullptr;
   }
 
-  GurumddsNodeInfo * node_info = static_cast<GurumddsNodeInfo *>(node->data);
-  if (node_info == nullptr) {
-    RMW_SET_ERROR_MSG("node info is null");
+  rmw_context_impl_t * ctx = node->context->impl;
+
+  rmw_publisher_t * const rmw_pub =
+    __rmw_create_publisher(
+    ctx,
+    node,
+    ctx->participant,
+    ctx->publisher,
+    type_supports,
+    topic_name,
+    qos_policies,
+    publisher_options,
+    ctx->localhost_only);
+
+  if (rmw_pub == nullptr) {
+    RMW_SET_ERROR_MSG("failed to create RMW publisher");
     return nullptr;
   }
 
-  dds_DomainParticipant * participant = node_info->participant;
-  if (participant == nullptr) {
-    RMW_SET_ERROR_MSG("participant handle is null");
-    return nullptr;
-  }
-
-  const rosidl_message_type_support_t * type_support =
-    get_message_typesupport_handle(type_supports, rosidl_typesupport_introspection_c__identifier);
-  if (type_support == nullptr) {
-    rcutils_reset_error();
-    type_support = get_message_typesupport_handle(
-      type_supports, rosidl_typesupport_introspection_cpp::typesupport_identifier);
-    if (type_support == nullptr) {
-      rcutils_reset_error();
-      RMW_SET_ERROR_MSG("type support not from this implementation");
-      return nullptr;
-    }
-  }
-
-  rmw_publisher_t * rmw_publisher = nullptr;
-  GurumddsPublisherInfo * publisher_info = nullptr;
-  dds_Publisher * dds_publisher = nullptr;
-  dds_PublisherQos publisher_qos;
-  dds_DataWriter * topic_writer = nullptr;
-  dds_DataWriterQos datawriter_qos;
-  dds_Topic * topic = nullptr;
-  dds_TopicDescription * topic_desc = nullptr;
-  dds_TypeSupport * dds_typesupport = nullptr;
-  dds_ReturnCode_t ret = dds_RETCODE_OK;
-  rmw_ret_t rmw_ret = RMW_RET_OK;
-
-  std::string type_name =
-    create_type_name(type_support->data, type_support->typesupport_identifier);
-  if (type_name.empty()) {
-    // Error message is already set
-    return nullptr;
-  }
-
-  std::string processed_topic_name;
-  if (!qos_policies->avoid_ros_namespace_conventions) {
-    processed_topic_name = std::string(ros_topic_prefix) + topic_name;
-  } else {
-    processed_topic_name = topic_name;
-  }
-
-  std::string metastring =
-    create_metastring(type_support->data, type_support->typesupport_identifier);
-  if (metastring.empty()) {
-    // Error message is already set
-    return nullptr;
-  }
-
-  dds_typesupport = dds_TypeSupport_create(metastring.c_str());
-  if (dds_typesupport == nullptr) {
-    RMW_SET_ERROR_MSG("failed to create typesupport");
-    goto fail;
-  }
-
-  ret = dds_TypeSupport_register_type(dds_typesupport, participant, type_name.c_str());
-  if (ret != dds_RETCODE_OK) {
-    RMW_SET_ERROR_MSG("failed to register type to domain participant");
-    goto fail;
-  }
-
-  ret = dds_DomainParticipant_get_default_publisher_qos(participant, &publisher_qos);
-  if (ret != dds_RETCODE_OK) {
-    RMW_SET_ERROR_MSG("failed to get default publisher qos");
-    goto fail;
-  }
-
-  dds_publisher = dds_DomainParticipant_create_publisher(participant, &publisher_qos, nullptr, 0);
-  if (dds_publisher == nullptr) {
-    RMW_SET_ERROR_MSG("failed to create publisher");
-    dds_PublisherQos_finalize(&publisher_qos);
-    goto fail;
-  }
-
-  ret = dds_PublisherQos_finalize(&publisher_qos);
-  if (ret != dds_RETCODE_OK) {
-    RMW_SET_ERROR_MSG("failed to finalize publisher qos");
-    goto fail;
-  }
-
-  // Look for message topic
-  topic_desc = dds_DomainParticipant_lookup_topicdescription(
-    participant, processed_topic_name.c_str());
-  if (topic_desc == nullptr) {
-    dds_TopicQos topic_qos;
-    ret = dds_DomainParticipant_get_default_topic_qos(participant, &topic_qos);
-    if (ret != dds_RETCODE_OK) {
-      RMW_SET_ERROR_MSG("failed to get default topic qos");
-      goto fail;
-    }
-
-    topic = dds_DomainParticipant_create_topic(
-      participant, processed_topic_name.c_str(), type_name.c_str(), &topic_qos, nullptr, 0);
-    if (topic == nullptr) {
-      RMW_SET_ERROR_MSG("failed to create topic");
-      dds_TopicQos_finalize(&topic_qos);
-      goto fail;
-    }
-
-    ret = dds_TopicQos_finalize(&topic_qos);
-    if (ret != dds_RETCODE_OK) {
-      RMW_SET_ERROR_MSG("failed to finalize topic qos");
-      goto fail;
-    }
-  } else {
-    dds_Duration_t timeout;
-    timeout.sec = 0;
-    timeout.nanosec = 1;
-    topic = dds_DomainParticipant_find_topic(participant, processed_topic_name.c_str(), &timeout);
-    if (topic == nullptr) {
-      RMW_SET_ERROR_MSG("failed to find topic");
-      goto fail;
-    }
-  }
-
-  if (!get_datawriter_qos(dds_publisher, qos_policies, &datawriter_qos)) {
-    // Error message already set
-    goto fail;
-  }
-
-  topic_writer = dds_Publisher_create_datawriter(dds_publisher, topic, &datawriter_qos, nullptr, 0);
-  if (topic_writer == nullptr) {
-    RMW_SET_ERROR_MSG("failed to create datawriter");
-    dds_DataWriterQos_finalize(&datawriter_qos);
-    goto fail;
-  }
-
-  ret = dds_DataWriterQos_finalize(&datawriter_qos);
-  if (ret != dds_RETCODE_OK) {
-    RMW_SET_ERROR_MSG("failed to finalize datawriter qos");
-    goto fail;
-  }
-
-  publisher_info = new(std::nothrow) GurumddsPublisherInfo();
-  if (publisher_info == nullptr) {
-    RMW_SET_ERROR_MSG("failed to allocate GurumddsPublisherInfo");
-    goto fail;
-  }
-
-  publisher_info->implementation_identifier = gurum_gurumdds_identifier;
-  publisher_info->publisher = dds_publisher;
-  publisher_info->topic_writer = topic_writer;
-  publisher_info->rosidl_message_typesupport = type_support;
-  publisher_info->publisher_gid.implementation_identifier = gurum_gurumdds_identifier;
-
-  static_assert(
-    sizeof(GurumddsPublisherGID) <= RMW_GID_STORAGE_SIZE,
-    "RMW_GID_STORAGE_SIZE insufficient to store the rmw_gurumdds_cpp GID implementation.");
-  memset(publisher_info->publisher_gid.data, 0, RMW_GID_STORAGE_SIZE);
-  {
-    auto publisher_gid =
-      reinterpret_cast<GurumddsPublisherGID *>(publisher_info->publisher_gid.data);
-    dds_DataWriter_get_guid(topic_writer, publisher_gid->publication_handle);
-  }
-
-  rmw_publisher = rmw_publisher_allocate();
-  if (rmw_publisher == nullptr) {
-    RMW_SET_ERROR_MSG("failed to allocate publisher");
-    goto fail;
-  }
-
-  rmw_publisher->implementation_identifier = gurum_gurumdds_identifier;
-  rmw_publisher->data = publisher_info;
-  rmw_publisher->topic_name = reinterpret_cast<const char *>(rmw_allocate(strlen(topic_name) + 1));
-  if (rmw_publisher->topic_name == nullptr) {
-    RMW_SET_ERROR_MSG("failed to allocate memory for topic name");
-    goto fail;
-  }
-  memcpy(const_cast<char *>(rmw_publisher->topic_name), topic_name, strlen(topic_name) + 1);
-  rmw_publisher->options = *publisher_options;
-  rmw_publisher->can_loan_messages = false;
-
-  rmw_ret = rmw_trigger_guard_condition(node_info->graph_guard_condition);
-  if (rmw_ret != RMW_RET_OK) {
-    // Error message already set
-    goto fail;
-  }
-
-  dds_TypeSupport_delete(dds_typesupport);
-  dds_typesupport = nullptr;
-
-  std::this_thread::sleep_for(std::chrono::milliseconds(5));
-
   RCUTILS_LOG_DEBUG_NAMED(
-    "rmw_gurumdds_cpp",
+    RMW_GURUMDDS_ID,
     "Created publisher with topic '%s' on node '%s%s%s'",
     topic_name, node->namespace_,
     node->namespace_[strlen(node->namespace_) - 1] == '/' ? "" : "/", node->name);
 
-  return rmw_publisher;
-
-fail:
-  if (rmw_publisher != nullptr) {
-    rmw_publisher_free(rmw_publisher);
-  }
-
-  if (topic != nullptr) {
-    dds_DomainParticipant_delete_topic(participant, topic);
-  }
-
-  if (dds_publisher != nullptr) {
-    if (topic_writer != nullptr) {
-      dds_Publisher_delete_datawriter(dds_publisher, topic_writer);
-    }
-    dds_DomainParticipant_delete_publisher(participant, dds_publisher);
-  }
-
-  if (dds_typesupport != nullptr) {
-    dds_TypeSupport_delete(dds_typesupport);
-  }
-
-  if (publisher_info != nullptr) {
-    delete publisher_info;
-  }
-
-  return nullptr;
+  return rmw_pub;
 }
 
 rmw_ret_t
@@ -333,29 +356,21 @@ rmw_publisher_count_matched_subscriptions(
   const rmw_publisher_t * publisher,
   size_t * subscription_count)
 {
-  if (publisher == nullptr) {
-    RMW_SET_ERROR_MSG("publisher handle is null");
-    return RMW_RET_INVALID_ARGUMENT;
-  }
+  RMW_CHECK_ARGUMENT_FOR_NULL(publisher, RMW_RET_INVALID_ARGUMENT);
+  RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
+    publisher,
+    publisher->implementation_identifier,
+    RMW_GURUMDDS_ID,
+    return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
+  RMW_CHECK_ARGUMENT_FOR_NULL(subscription_count, RMW_RET_INVALID_ARGUMENT);
 
-  if (subscription_count == nullptr) {
-    RMW_SET_ERROR_MSG("subscription_count is null");
-    return RMW_RET_INVALID_ARGUMENT;
-  }
-
-  auto info = static_cast<GurumddsPublisherInfo *>(publisher->data);
-  if (info == nullptr) {
+  auto publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
+  if (publisher_info == nullptr) {
     RMW_SET_ERROR_MSG("publisher internal data is invalid");
     return RMW_RET_ERROR;
   }
 
-  dds_Publisher * dds_publisher = info->publisher;
-  if (dds_publisher == nullptr) {
-    RMW_SET_ERROR_MSG("dds publisher is null");
-    return RMW_RET_ERROR;
-  }
-
-  dds_DataWriter * topic_writer = info->topic_writer;
+  dds_DataWriter * topic_writer = publisher_info->topic_writer;
   if (topic_writer == nullptr) {
     RMW_SET_ERROR_MSG("topic writer is null");
     return RMW_RET_ERROR;
@@ -367,7 +382,7 @@ rmw_publisher_count_matched_subscriptions(
     dds_InstanceHandleSeq_delete(seq);
     return RMW_RET_ERROR;
   }
-  *subscription_count = (size_t)dds_InstanceHandleSeq_length(seq);
+  *subscription_count = static_cast<size_t>(dds_InstanceHandleSeq_length(seq));
 
   dds_InstanceHandleSeq_delete(seq);
 
@@ -379,18 +394,18 @@ rmw_publisher_assert_liveliness(const rmw_publisher_t * publisher)
 {
   RMW_CHECK_ARGUMENT_FOR_NULL(publisher, RMW_RET_INVALID_ARGUMENT);
 
-  auto * info = static_cast<GurumddsPublisherInfo *>(publisher->data);
-  if (info == nullptr) {
+  auto publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
+  if (publisher_info == nullptr) {
     RMW_SET_ERROR_MSG("publisher internal data is invalid");
     return RMW_RET_ERROR;
   }
 
-  if (info->topic_writer == nullptr) {
+  if (publisher_info->topic_writer == nullptr) {
     RMW_SET_ERROR_MSG("publisher internal datawriter is invalid");
     return RMW_RET_ERROR;
   }
 
-  if (dds_DataWriter_assert_liveliness(info->topic_writer) != dds_RETCODE_OK) {
+  if (dds_DataWriter_assert_liveliness(publisher_info->topic_writer) != dds_RETCODE_OK) {
     RMW_SET_ERROR_MSG("failed to assert liveliness of datawriter");
     return RMW_RET_ERROR;
   }
@@ -404,10 +419,11 @@ rmw_publisher_wait_for_all_acked(const rmw_publisher_t * publisher, rmw_time_t w
   RMW_CHECK_ARGUMENT_FOR_NULL(publisher, RMW_RET_INVALID_ARGUMENT);
   RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
     publisher,
-    publisher->implementation_identifier, gurum_gurumdds_identifier,
+    publisher->implementation_identifier,
+    RMW_GURUMDDS_ID,
     return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
 
-  GurumddsPublisherInfo * publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
+  auto publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
   if (publisher_info == nullptr) {
     RMW_SET_ERROR_MSG("publisher internal data is invalid");
     return RMW_RET_ERROR;
@@ -431,71 +447,43 @@ rmw_destroy_publisher(rmw_node_t * node, rmw_publisher_t * publisher)
 {
   RMW_CHECK_ARGUMENT_FOR_NULL(node, RMW_RET_INVALID_ARGUMENT);
   RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
-    node handle,
-    node->implementation_identifier, gurum_gurumdds_identifier,
+    node,
+    node->implementation_identifier,
+    RMW_GURUMDDS_ID,
     return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
 
   RMW_CHECK_ARGUMENT_FOR_NULL(publisher, RMW_RET_INVALID_ARGUMENT);
   RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
-    publisher handle,
-    publisher->implementation_identifier, gurum_gurumdds_identifier,
+    publisher,
+    publisher->implementation_identifier,
+    RMW_GURUMDDS_ID,
     return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
 
-  auto node_info = static_cast<GurumddsNodeInfo *>(node->data);
-  if (node_info == nullptr) {
-    RMW_SET_ERROR_MSG("node info handle is null");
-    return RMW_RET_ERROR;
-  }
-  auto participant = static_cast<dds_DomainParticipant *>(node_info->participant);
-  if (participant == nullptr) {
-    RMW_SET_ERROR_MSG("participant handle is null");
+  rmw_context_impl_t * ctx = node->context->impl;
+
+  if (graph_on_publisher_deleted(
+      ctx, node, reinterpret_cast<GurumddsPublisherInfo *>(publisher->data)))
+  {
+    RCUTILS_LOG_ERROR_NAMED(RMW_GURUMDDS_ID, "failed to update graph for publisher");
     return RMW_RET_ERROR;
   }
 
-  GurumddsPublisherInfo * publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
-  dds_ReturnCode_t ret = dds_RETCODE_OK;
-  if (publisher_info) {
-    dds_Publisher * dds_publisher = publisher_info->publisher;
+  rmw_ret_t ret = __rmw_destroy_publisher(ctx, publisher);
 
-    if (dds_publisher != nullptr) {
-      if (publisher_info->topic_writer != nullptr) {
-        ret = dds_Publisher_delete_datawriter(dds_publisher, publisher_info->topic_writer);
-        if (ret != dds_RETCODE_OK) {
-          RMW_SET_ERROR_MSG("failed to delete datawriter");
-          return RMW_RET_ERROR;
-        }
-        publisher_info->topic_writer = nullptr;
-      }
-
-      ret = dds_DomainParticipant_delete_publisher(participant, dds_publisher);
-      if (ret != dds_RETCODE_OK) {
-        RMW_SET_ERROR_MSG("failed to delete publisher");
-        return RMW_RET_ERROR;
-      }
-      publisher_info->publisher = nullptr;
-    } else if (publisher_info->topic_writer != nullptr) {
-      RMW_SET_ERROR_MSG("cannot delete datawriter because the publisher is null");
-      return RMW_RET_ERROR;
-    }
-
-    delete publisher_info;
-    publisher->data = nullptr;
+  if (ret == RMW_RET_OK) {
     if (publisher->topic_name != nullptr) {
       RCUTILS_LOG_DEBUG_NAMED(
-        "rmw_gurumdds_cpp",
+        RMW_GURUMDDS_ID,
         "Deleted publisher with topic '%s' on node '%s%s%s'",
         publisher->topic_name, node->namespace_,
         node->namespace_[strlen(node->namespace_) - 1] == '/' ? "" : "/", node->name);
 
       rmw_free(const_cast<char *>(publisher->topic_name));
     }
+    rmw_publisher_free(publisher);
   }
 
-  rmw_publisher_free(publisher);
-
-  rmw_ret_t rmw_ret = rmw_trigger_guard_condition(node_info->graph_guard_condition);
-
-  return rmw_ret;
+  return ret;
 }
 
 rmw_ret_t
@@ -504,19 +492,19 @@ rmw_get_gid_for_publisher(const rmw_publisher_t * publisher, rmw_gid_t * gid)
   RMW_CHECK_ARGUMENT_FOR_NULL(publisher, RMW_RET_INVALID_ARGUMENT);
   RMW_CHECK_ARGUMENT_FOR_NULL(gid, RMW_RET_INVALID_ARGUMENT);
   RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
-    publisher handle,
+    publisher,
     publisher->implementation_identifier,
-    gurum_gurumdds_identifier,
+    RMW_GURUMDDS_ID,
     return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
 
-  const GurumddsPublisherInfo * info =
-    static_cast<const GurumddsPublisherInfo *>(publisher->data);
-  if (info == nullptr) {
+  auto publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
+  if (publisher_info == nullptr) {
     RMW_SET_ERROR_MSG("publisher info handle is null");
     return RMW_RET_ERROR;
   }
 
-  *gid = info->publisher_gid;
+  *gid = publisher_info->publisher_gid;
+
   return RMW_RET_OK;
 }
 
@@ -527,33 +515,38 @@ rmw_publisher_get_actual_qos(
 {
   RMW_CHECK_ARGUMENT_FOR_NULL(publisher, RMW_RET_INVALID_ARGUMENT);
   RMW_CHECK_ARGUMENT_FOR_NULL(qos, RMW_RET_INVALID_ARGUMENT);
+  RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
+    publisher,
+    publisher->implementation_identifier,
+    RMW_GURUMDDS_ID,
+    return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
 
-  GurumddsPublisherInfo * info = static_cast<GurumddsPublisherInfo *>(publisher->data);
-  if (info == nullptr) {
+  auto publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
+  if (publisher_info == nullptr) {
     RMW_SET_ERROR_MSG("publisher internal data is invalid");
     return RMW_RET_ERROR;
   }
 
-  dds_DataWriter * data_writer = info->topic_writer;
-  if (data_writer == nullptr) {
+  dds_DataWriter * topic_writer = publisher_info->topic_writer;
+  if (topic_writer == nullptr) {
     RMW_SET_ERROR_MSG("publisher internal data writer is invalid");
     return RMW_RET_ERROR;
   }
 
   dds_DataWriterQos dds_qos;
-  dds_ReturnCode_t ret = dds_DataWriter_get_qos(data_writer, &dds_qos);
+  dds_ReturnCode_t ret = dds_DataWriter_get_qos(topic_writer, &dds_qos);
   if (ret != dds_RETCODE_OK) {
     RMW_SET_ERROR_MSG("publisher can't get data writer qos policies");
     return RMW_RET_ERROR;
   }
 
-  qos->reliability = convert_reliability(dds_qos.reliability);
-  qos->durability = convert_durability(dds_qos.durability);
-  qos->deadline = convert_deadline(dds_qos.deadline);
-  qos->lifespan = convert_lifespan(dds_qos.lifespan);
-  qos->liveliness = convert_liveliness(dds_qos.liveliness);
-  qos->liveliness_lease_duration = convert_liveliness_lease_duration(dds_qos.liveliness);
-  qos->history = convert_history(dds_qos.history);
+  qos->reliability = convert_reliability(&dds_qos.reliability);
+  qos->durability = convert_durability(&dds_qos.durability);
+  qos->deadline = convert_deadline(&dds_qos.deadline);
+  qos->lifespan = convert_lifespan(&dds_qos.lifespan);
+  qos->liveliness = convert_liveliness(&dds_qos.liveliness);
+  qos->liveliness_lease_duration = convert_liveliness_lease_duration(&dds_qos.liveliness);
+  qos->history = convert_history(&dds_qos.history);
   qos->depth = static_cast<size_t>(dds_qos.history.depth);
 
   ret = dds_DataWriterQos_finalize(&dds_qos);
@@ -572,25 +565,22 @@ rmw_publish(
   rmw_publisher_allocation_t * allocation)
 {
   (void)allocation;
-  RCUTILS_CHECK_FOR_NULL_WITH_MSG(
-    publisher, "publisher pointer is null",
-    return RMW_RET_INVALID_ARGUMENT);
-  RCUTILS_CHECK_FOR_NULL_WITH_MSG(
-    ros_message, "ros_message pointer is null",
-    return RMW_RET_INVALID_ARGUMENT);
+  RMW_CHECK_ARGUMENT_FOR_NULL(publisher, RMW_RET_INVALID_ARGUMENT);
+  RMW_CHECK_ARGUMENT_FOR_NULL(ros_message, RMW_RET_INVALID_ARGUMENT);
   RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
     publisher,
     publisher->implementation_identifier,
-    gurum_gurumdds_identifier,
+    RMW_GURUMDDS_ID,
     return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
 
-  auto info = static_cast<GurumddsPublisherInfo *>(publisher->data);
-  RCUTILS_CHECK_FOR_NULL_WITH_MSG(info, "publisher info pointer is null", return RMW_RET_ERROR);
+  auto publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
+  RCUTILS_CHECK_ARGUMENT_FOR_NULL(publisher_info, RMW_RET_ERROR);
 
-  dds_DataWriter * topic_writer = info->topic_writer;
-  RCUTILS_CHECK_FOR_NULL_WITH_MSG(info, "topic writer is null", return RMW_RET_ERROR);
+  dds_DataWriter * topic_writer = publisher_info->topic_writer;
+  RCUTILS_CHECK_ARGUMENT_FOR_NULL(topic_writer, RMW_RET_ERROR);
 
-  const rosidl_message_type_support_t * rosidl_typesupport = info->rosidl_message_typesupport;
+  const rosidl_message_type_support_t * rosidl_typesupport =
+    publisher_info->rosidl_message_typesupport;
   if (rosidl_typesupport == nullptr) {
     RMW_SET_ERROR_MSG("rosidl typesupport handle is null");
     return RMW_RET_ERROR;
@@ -622,7 +612,16 @@ rmw_publish(
     return RMW_RET_ERROR;
   }
 
-  dds_ReturnCode_t ret = dds_DataWriter_raw_write(topic_writer, dds_message, size);
+  dds_SampleInfoEx sampleinfo_ex;
+  memset(&sampleinfo_ex, 0, sizeof(dds_SampleInfoEx));
+  ros_sn_to_dds_sn(++publisher_info->sequence_number, &sampleinfo_ex.seq);
+  ros_guid_to_dds_guid(
+    publisher_info->publisher_gid.data,
+    reinterpret_cast<uint8_t *>(&sampleinfo_ex.src_guid));
+
+  dds_ReturnCode_t ret = dds_DataWriter_raw_write_w_sampleinfoex(
+    topic_writer, dds_message, size, &sampleinfo_ex);
+
   const char * errstr;
   if (ret == dds_RETCODE_OK) {
     errstr = "dds_RETCODE_OK";
@@ -641,8 +640,8 @@ rmw_publish(
     free(dds_message);
     return RMW_RET_ERROR;
   }
-  const char * topic_name = dds_Topic_get_name(dds_DataWriter_get_topic(topic_writer));
-  RCUTILS_LOG_DEBUG_NAMED("rmw_gurumdds_cpp", "Published data on topic %s", topic_name);
+
+  RCUTILS_LOG_DEBUG_NAMED(RMW_GURUMDDS_ID, "Published data on topic %s", publisher->topic_name);
 
   free(dds_message);
 
@@ -656,29 +655,34 @@ rmw_publish_serialized_message(
   rmw_publisher_allocation_t * allocation)
 {
   (void)allocation;
-  RCUTILS_CHECK_FOR_NULL_WITH_MSG(
-    publisher, "publisher pointer is null",
-    return RMW_RET_INVALID_ARGUMENT);
-  RCUTILS_CHECK_FOR_NULL_WITH_MSG(
-    serialized_message, "serialized_message pointer is null",
-    return RMW_RET_INVALID_ARGUMENT);
+  RMW_CHECK_ARGUMENT_FOR_NULL(publisher, RMW_RET_INVALID_ARGUMENT);
+  RMW_CHECK_ARGUMENT_FOR_NULL(serialized_message, RMW_RET_INVALID_ARGUMENT);
   RMW_CHECK_TYPE_IDENTIFIERS_MATCH(
     publisher,
     publisher->implementation_identifier,
-    gurum_gurumdds_identifier,
+    RMW_GURUMDDS_ID,
     return RMW_RET_INCORRECT_RMW_IMPLEMENTATION);
 
-  auto info = static_cast<GurumddsPublisherInfo *>(publisher->data);
-  RCUTILS_CHECK_FOR_NULL_WITH_MSG(info, "publisher info pointer is null", return RMW_RET_ERROR);
+  auto publisher_info = static_cast<GurumddsPublisherInfo *>(publisher->data);
+  RCUTILS_CHECK_ARGUMENT_FOR_NULL(publisher_info, RMW_RET_ERROR);
 
-  dds_DataWriter * topic_writer = info->topic_writer;
-  RCUTILS_CHECK_FOR_NULL_WITH_MSG(info, "topic writer is null", return RMW_RET_ERROR);
+  dds_DataWriter * topic_writer = publisher_info->topic_writer;
+  RCUTILS_CHECK_ARGUMENT_FOR_NULL(topic_writer, RMW_RET_ERROR);
 
-  dds_ReturnCode_t ret = dds_DataWriter_raw_write(
+  dds_SampleInfoEx sampleinfo_ex;
+  memset(&sampleinfo_ex, 0, sizeof(dds_SampleInfoEx));
+  ros_sn_to_dds_sn(++publisher_info->sequence_number, &sampleinfo_ex.seq);
+  ros_guid_to_dds_guid(
+    publisher_info->publisher_gid.data,
+    reinterpret_cast<uint8_t *>(&sampleinfo_ex.src_guid));
+
+  dds_ReturnCode_t ret = dds_DataWriter_raw_write_w_sampleinfoex(
     topic_writer,
     serialized_message->buffer,
-    static_cast<uint32_t>(serialized_message->buffer_length)
+    static_cast<uint32_t>(serialized_message->buffer_length),
+    &sampleinfo_ex
   );
+
   const char * errstr;
   if (ret == dds_RETCODE_OK) {
     errstr = "dds_RETCODE_OK";
@@ -696,8 +700,8 @@ rmw_publish_serialized_message(
     RMW_SET_ERROR_MSG(errmsg.str().c_str());
     return RMW_RET_ERROR;
   }
-  const char * topic_name = dds_Topic_get_name(dds_DataWriter_get_topic(topic_writer));
-  RCUTILS_LOG_DEBUG_NAMED("rmw_gurumdds_cpp", "Published data on topic %s", topic_name);
+
+  RCUTILS_LOG_DEBUG_NAMED(RMW_GURUMDDS_ID, "Published data on topic %s", publisher->topic_name);
 
   return RMW_RET_OK;
 }
